from typing import Dict

from ..base import APIObject, APIList, Session
from ..kv import KV

from .. import users
from .. import apps
from ..notifications import Notifications

from . import registry


class Object(APIObject):
    props = {
        "name",
        "description",
        "icon",
        "access",
        "meta",
        "tags",
        "key",
        "owner_scope",
    }

    def __init__(self, objectData: Dict, session: Session):
        super().__init__(
            f"api/objects/{objectData['id']}",
            {"object": objectData["id"]},
            session,
            cached_data=objectData,
        )
        # The object ID
        self.id = objectData["id"]

        self._kv = KV(f"api/kv/objects/{objectData['id']}", self.session)

    @property
    def kv(self):
        return self._kv

    @kv.setter
    def kv(self, v):
        return self._kv.set(**v)

    def __getattr__(self, attr):
        try:
            return self.cached_data[attr]
        except:
            return None

    @property
    def owner(self):
        return users.User(self.cached_data["owner"], self.session)

    @property
    def app(self):
        if self.cached_data["app"] is None:
            return None
        return apps.App(self.cached_data["app"], session=self.session)

    def update(self, **kwargs):
        """
        Updates the given data::

            o.update(name="My new name",description="my new description")
        """

        meta = self.cached_data.get("meta", None)

        def updateMeta(o):
            if "result" in o and o["result"] == "ok" and "meta" in kwargs:
                if meta is None:
                    self.cached_data.pop("meta", 0)
                else:
                    # We have values of meta, so set them correctly
                    kwm = kwargs["meta"]
                    for key in kwm:
                        if kwm[key] is None:
                            meta.pop(key, 0)
                        else:
                            meta[key] = kwm[key]
                    self.cached_data["meta"] = meta
            return o

        return self.session.f(super().update(**kwargs), updateMeta)


class Objects(APIList):
    def __init__(self, constraints: Dict, session: Session):
        super().__init__("api/objects", constraints, session)

    def __getitem__(self, item):
        return super()._getitem(item, f=lambda x: registry.getObject(x, self.session))

    def __call__(self, **kwargs):
        # To query by object type, we use _type
        if "_type" in kwargs:
            kwargs["type"] = kwargs["_type"]
            del kwargs["_type"]
        return super()._call(
            f=lambda x: [registry.getObject(xx, self.session) for xx in x], **kwargs
        )

    def create(self, name, meta={}, _type="timeseries", **kwargs):
        """
        Creates a new object of the given type (timeseries by default).
        """
        return super()._create(
            f=lambda x: registry.getObject(x, self.session),
            **{"name": name, "type": _type, "meta": meta, **kwargs},
        )



import pytest
from heedy import App

"""Dashboard has been temporarily disabled
def lists_equal(l1, l2):
    assert len(l1) == len(l2)
    for i in range(len(l1)):
        assert l1[i] == l2[i]


def test_dashboard():
    a = App("testkey")
    t1 = a.objects.create("obj1", key="key1")
    t2 = a.objects.create("obj2", key="key2")
    t3 = a.objects.create("obj3", key="key3")

    t1.insert_array(
        [{"t": 1, "d": 1}, {"t": 2, "d": 1},]
    )
    t2.insert_array(
        [
            {"t": 1.1, "d": 2},
            {"t": 2.1, "d": 2},
            {"t": 3.1, "d": 2},
            {"t": 4.1, "d": 2},
            {"t": 5.1, "d": 2},
        ]
    )
    t3.insert_array(
        [
            {"t": 1.2, "d": 3},
            {"t": 2.2, "d": 3},
            {"t": 3.3, "d": 3},
            {"t": 4.4, "d": 3},
            {"t": 5.5, "d": 3},
        ]
    )
    d = a.objects.create("db", otype="dashboard")

    d.add(
        [
            {
                "merge": [
                    {"timeseries": t1.id},
                    {"timeseries": t2.id, "t1": 3.0},
                    {"timeseries": t3.id, "i1": 1, "i2": 2},
                ]
            }
        ]
    )

    elements = d[:]
    assert len(elements) == 1

    assert elements[0].title == ""
    elements[0].title = "Hello World"
    assert elements[0].title == "Hello World"
    assert elements[0].settings == {}
    elements[0].settings = {"hi": "ho"}
    assert elements[0].settings == {"hi": "ho"}

    lists_equal(
        elements[0].data[0],
        [
            {"t": 1, "d": 1},
            {"t": 2, "d": 1},
            {"t": 2.2, "d": 3},
            {"t": 3.1, "d": 2},
            {"t": 4.1, "d": 2},
            {"t": 5.1, "d": 2},
        ],
    )

    t1.insert_array(
        [{"t": 3, "d": 1}, {"t": 4, "d": 1}, {"t": 5, "d": 1},]
    )

    element = d[elements[0].id]

    lists_equal(
        element.data[0],
        [
            {"t": 1, "d": 1},
            {"t": 2, "d": 1},
            {"t": 2.2, "d": 3},
            {"t": 3, "d": 1},
            {"t": 3.1, "d": 2},
            {"t": 4, "d": 1},
            {"t": 4.1, "d": 2},
            {"t": 5, "d": 1},
            {"t": 5.1, "d": 2},
        ],
    )

    for o in a.objects():
        o.delete()


def test_dashboard_indices():
    a = App("testkey")
    t = a.objects.create("obj1", key="key1")
    t.insert_array(
        [{"t": 1, "d": 1}, {"t": 2, "d": 1},]
    )
    d = a.objects.create("db", otype="dashboard")

    d.add([{"timeseries": t.id}], title="lol")
    d.add([{"timeseries": t.id}], title="lol2")
    d.add([{"timeseries": t.id}], title="lol3")
    d.add([{"timeseries": t.id}], title="lol4")

    els = d[:]
    assert len(els) == 4
    assert els[0].title == "lol"
    assert els[1]["title"] == "lol2"
    assert els[2]["title"] == "lol3"
    assert els[3]["title"] == "lol4"

    els[2].index = 1  # switch lol2 and lol3

    els = d[:]
    assert len(els) == 4
    assert els[0].title == "lol" and els[0].index == 0
    assert els[1]["title"] == "lol3" and els[1]["index"] == 1
    assert els[2]["title"] == "lol2" and els[2]["index"] == 2
    assert els[3]["title"] == "lol4" and els[3]["index"] == 3

    els[1].index = 2  # switch back

    els = d[:]
    assert len(els) == 4
    assert els[0].title == "lol" and els[0].index == 0
    assert els[1]["title"] == "lol2" and els[1]["index"] == 1
    assert els[2]["title"] == "lol3" and els[2]["index"] == 2
    assert els[3]["title"] == "lol4" and els[3]["index"] == 3

    els[1].delete()

    els = d[:]
    assert len(els) == 3
    assert els[0].title == "lol" and els[0].index == 0
    assert els[1]["title"] == "lol3" and els[1]["index"] == 1
    assert els[2]["title"] == "lol4" and els[2]["index"] == 2

    els[2].delete()

    els = d[:]
    assert len(els) == 2
    assert els[0].title == "lol" and els[0].index == 0
    assert els[1]["title"] == "lol3" and els[1]["index"] == 1

    els[0].delete()
    els[1].delete()

    els = d[:]
    assert len(els) == 0

    for o in a.objects():
        o.delete()
"""


from ..base import APIObject, Session
from .objects import Object
from .registry import registerObjectType


class DashboardElement(APIObject):
    props = {"query", "settings", "data", "id", "type", "title", "on_demand", "index"}

    def __init__(self, dashboard, cached_data={}):

        super().__init__(
            f"api/objects/{dashboard.id}/dashboard/{cached_data['id']}",
            {},
            dashboard.session,
            cached_data,
        )

        # Dashboard elements don't have notifications
        del self.notifications


class Dashboard(Object):
    def __getitem__(self, getrange):
        """Allows accessing the dashboard's elements as if it were a big python array
        """
        if isinstance(getrange, str):
            # The item is an ID, so return the dashboard element
            return self.session.get(
                self.uri + "/dashboard/" + getrange,
                f=lambda x: DashboardElement(self, x),
            )

        return self.session.get(
            self.uri + "/dashboard",
            f=lambda x: [DashboardElement(self, xx) for xx in x][getrange],
        )

    def __setitem__(self, elementid, value):
        return self.session.post(self.uri + "/dashboard/" + elementid, value)

    def add(self, query, etype="dataset", **kwargs):
        kwargs["type"] = etype
        kwargs["query"] = query
        return self.session.post(self.uri + "/dashboard", [kwargs])


registerObjectType("dashboard", Dashboard)




from .base import Session
from typing import Dict


class Notifications:
    def __init__(self, constraints: Dict, session: Session):
        self.constraints = constraints
        self.session = session

    def __call__(self, **kwargs):
        return self.session.get(
            "api/notifications", params={**self.constraints, **kwargs}
        )

    def notify(self, key, title=None, **kwargs):
        n = {**self.constraints, **{"key": key}, **kwargs}
        if title is not None:
            n["title"] = title
        if "_global" in n:
            n["global"] = n["_global"]
            del n["_global"]
        if "_type" in n:
            n["type"] = n["_type"]
            del n["_type"]
        return self.session.post("api/notifications", n)

    def delete(self, key=None, **kwargs):
        if key is not None:
            kwargs["key"] = key
        return self.session.delete(
            "/api/notifications", params={**self.constraints, **kwargs}
        )



from .objects.timeseries import fixTimestamps, Timeseries, DatapointArray


def getQuery(ts, kwargs):
    fixTimestamps(kwargs)
    if isinstance(ts, Merge):
        kwargs["merge"] = ts.query
        return kwargs
    if isinstance(ts, Timeseries):
        ts = ts.id
    elif not isinstance(ts, str):
        raise Exception("Timeseries must be either ID string or a Timeseries object")
    kwargs["timeseries"] = ts
    return kwargs


class Merge:
    """Merge represents a query which allows to merge multiple timeseries into one
    when reading, with all the data merged together by increasing timestamp.
    The merge query is used as a constructor-type object::
        m = Merge(h)
        m.add("objectid",t1="now-10h")
        m.add(h.objects["objectid2"],t1="now-10h")
        result = m.run()
    """

    def __init__(self, h):
        """Given a heedy connection, begins the construction of a merge query"""
        self.h = h
        self.query = []

    def add(self, ts, **kwargs):
        self.query.append(getQuery(ts, kwargs))

    def run(self):
        return self.h.session.post(
            "/api/timeseries/dataset",
            {"data": {"merge": self.query}},
            f=lambda x: DatapointArray(x["data"]),
        )


class Dataset(object):
    """Heedy is capable of taking several separate unrelated timeseries, and based upon
    the chosen interpolation method, putting them all together to generate tabular data centered about
    either another timeseries' datapoints, or based upon time intervals.
    The underlying issue that Datasets solve is that in Heedy, timeseries are inherently unrelated.
    In most data stores, such as standard relational (SQL) databases, and even excel spreadsheets, data is in tabular
    form. That is, if we have measurements of temperature in our house and our mood, we have a table:
        +--------------+----------------------+
        | Mood Rating  | Room Temperature (F) |
        +==============+======================+
        | 7            | 73                   |
        +--------------+----------------------+
        | 3            | 84                   |
        +--------------+----------------------+
        | 5            | 79                   |
        +--------------+----------------------+
    The benefit of having such a table is that it is easy to perform data analysis. You know which temperature
    value corresponds to which mood rating. The downside of having such tables
    is that Mood Rating and Room Temperature must be directly related - a temperature measurement must be made
    each time a mood rating is given. Heedy has no such restrictions. Mood Rating and Room Temperature
    can be entirely separate sensors, which update data at their own rate. In Heedy, each timeseries
    can be inserted with any timestamp, and without regard for any other data.
    This separation of Timeseries makes data require some preprocessing and interpolation before it can be used
    for analysis. This is the purpose of the Dataset query. Heedy can put several streams together based
    upon chosen transforms and interpolators, returning a tabular structure which can readily be used for ML
    and statistical applications.
    There are two types of dataset queries
    :T-Dataset:
        T-Dataset: A dataset query which is generated based upon a time range. That is, you choose a time range and a
        time difference between elements of the dataset, and that is used to generate your dataset.
            +--------------+----------------------+
            | Timestamp    | Room Temperature (F) |
            +==============+======================+
            | 1pm          | 73                   |
            +--------------+----------------------+
            | 4pm          | 84                   |
            +--------------+----------------------+
            | 8pm          | 79                   |
            +--------------+----------------------+
        If I were to generate a T-dataset from 12pm to 8pm with dt=2 hours, using the interpolator "closest",
        I would get the following result:
            +--------------+----------------------+
            | Timestamp    | Room Temperature (F) |
            +==============+======================+
            | 12pm         | 73                   |
            +--------------+----------------------+
            | 2pm          | 73                   |
            +--------------+----------------------+
            | 4pm          | 84                   |
            +--------------+----------------------+
            | 6pm          | 84                   |
            +--------------+----------------------+
            | 8pm          | 79                   |
            +--------------+----------------------+
        The "closest" interpolator happens to return the datapoint closest to the given timestamp. There are many
        interpolators to choose from (described later).
        Hint: T-Datasets can be useful for plotting data (such as daily or weekly averages).
    :X-Dataset:
        X-datasets allow to generate datasets based not on evenly spaced timestamps, but based upon values of a timeseries.
        Suppose you have the following data:
            +-----------+--------------+---+-----------+----------------------+
            | Timestamp | Mood Rating  |   | Timestamp | Room Temperature (F) |
            +===========+==============+===+===========+======================+
            | 1pm       | 7            |   | 2pm       | 73                   |
            +-----------+--------------+---+-----------+----------------------+
            | 4pm       | 3            |   | 5pm       | 84                   |
            +-----------+--------------+---+-----------+----------------------+
            | 11pm      | 5            |   | 8pm       | 81                   |
            +-----------+--------------+---+-----------+----------------------+
            |           |              |   | 11pm      | 79                   |
            +-----------+--------------+---+-----------+----------------------+
        An X-dataset with X=Mood Rating, and the interpolator "closest" on Room Temperature would generate:
            +--------------+----------------------+
            | Mood Rating  | Room Temperature (F) |
            +==============+======================+
            | 7            | 73                   |
            +--------------+----------------------+
            | 3            | 84                   |
            +--------------+----------------------+
            | 5            | 79                   |
            +--------------+----------------------+
    :Interpolators:
        Interpolators are special functions which specify how exactly the data is supposed to be combined
        into a dataset. Any PipeScript script can be used as an interpolator, including "sum", "count" and other transforms.
        By default, the "closest" interpolator is used, which simply returns the datapoint closest to the reference timestamp.
    """

    def __init__(self, h, x=None, **kwargs):
        """In order to begin dataset generation, you need to specify the reference time range or stream.
        To generate a T-dataset::
            d = Dataset(h, t1=start, t2=end, dt=tchange)
        To generate an X-dataset::
            d = Dataset(h,"mytimeseriesid", t1=start, t2=end)
        Note that everywhere you insert a timeseries ID, you are also free to insert Timeseries objects
        or even Merge queries. The Dataset query in Heedy supports merges natively for each field.
        """
        self.h = h

        if x is not None:
            if "dt" in kwargs:
                raise Exception(
                    "Can't do both T-dataset and X-dataset at the same time"
                )
            self.query = getQuery(x, kwargs)
        elif "dt" in kwargs:
            fixTimestamps(kwargs)
            self.query = kwargs
        else:
            raise Exception("Dataset must have either x or dt parameter")

        self.query["dataset"] = {}

    def add(self, key, ts, **kwargs):
        """Adds the given timeseries to the query construction. Unless an interpolator is specified, "closest" will be used.
        You can insert a merge query instead of a timeseries::
            d = Dataset(h, t1="now-1h",t2="now",dt=10)
            d.add("temperature","timeseriesid")
            d.add("steps",h.objects["timeseriesid2"])
            m = Merge(h)
            m.add("id1")
            m.add("id2")
            d.add("altitude",m)
            result = d.run()
        """

        if key in self.query["dataset"]:
            raise Exception("The key already exists")

        self.query["dataset"][key] = getQuery(ts, kwargs)

    def run(self):
        """Runs the dataset query, and returns the result"""
        return self.h.session.post(
            "/api/timeseries/dataset",
            {"data": self.query},
            f=lambda x: DatapointArray(x["data"]),
        )



import pytest
from heedy import App


def test_ts():
    a = App("testkey")
    ts = a.objects.create("myts", key="key1")
    assert len(ts) == 0
    ts.insert("hi!")
    assert len(ts) == 1
    assert len(ts(t1="now-10s")) == 1
    assert len(ts(t1="now")) == 0



import pytest
from heedy import App


def test_basics():
    a = App("testkey")
    assert a.owner.username == "test"

    a.owner.name = "Myname"
    assert a.owner.name == "Myname"

    assert len(a.objects()) == 0

    o = a.objects.create("myobj", {"schema": {"type": "number"}})
    assert o.name == "myobj"
    assert o.type == "timeseries"
    assert len(a.objects()) == 1

    assert o == a.objects[o.id]

    assert o.length() == 0
    o.append(2)
    assert o.length() == 1
    d = o[:]
    assert len(d) == 1
    assert d[0]["d"] == 2
    assert "dt" not in d[0]
    o.append(3, duration=9)
    d = o[:]
    assert len(d) == 2
    assert d[1]["d"] == 3
    assert d[1]["dt"] == 9
    o.remove()  # Clear the timeseries
    assert len(o) == 0

    o.delete()
    assert len(a.objects()) == 0
    # assert len(a.owner.apps())==1


def test_notifications():
    a = App("testkey")

    assert len(a.notifications()) == 0
    a.notify("hi", "hello")
    assert len(a.notifications()) == 1
    a.notifications.delete("hi")
    assert len(a.notifications()) == 0


def test_kv():
    a = App("testkey")

    assert len(a.kv()) == 0
    a.kv["test"] = True
    assert len(a.kv()) == 1
    assert a.kv["test"] == True

    del a.kv["test"]
    assert len(a.kv()) == 0


def test_tags_and_keys():
    a = App("testkey")
    a.objects.create("obj1", tags="tag1 tag2", key="key1")
    a.objects.create("obj2", tags="tag1 tag3", key="key2")

    assert len(a.objects(key="key1")) == 1
    assert len(a.objects(key="key2")) == 1
    assert len(a.objects(key="key3")) == 0
    assert len(a.objects(key="")) == 0

    assert len(a.objects(tags="tag1")) == 2
    assert len(a.objects(tags="tag1 tag3")) == 1
    assert len(a.objects(tags="tag1 tag3", key="key1")) == 0

    with pytest.raises(Exception):
        a.objects.create("obj3", tags="tag1 tag3", key="key2")

    a.objects(key="key1")[0].tags = "tag4"
    assert len(a.objects(tags="tag1")) == 1
    assert len(a.objects(tags="tag4")) == 1

    a.objects(key="key2")[0].key = ""
    assert len(a.objects(key="key2")) == 0
    assert len(a.objects(key="")) == 1

    for o in a.objects():
        o.delete()


def test_appschema():
    a = App("testkey")

    with pytest.raises(Exception):
        a.settings = {"lol": 12}

    a.settings_schema = {"lol": {"type": "number", "default": 42}}
    assert a.settings["lol"] == 42

    a.settings = {"lol": 12}
    assert a.settings["lol"] == 12

    with pytest.raises(Exception):
        a.settings = {"lol": "hi"}
    with pytest.raises(Exception):
        a.settings = {"lol": 24, "hee": 1}

    with pytest.raises(Exception):
        a.settings_schema = {"lol": {"type": "string", "default": "hi"}}

    a.update(
        settings={"lol": "hi"},
        settings_schema={"lol": {"type": "string", "default": "hello"}},
    )
    assert a.settings["lol"] == "hi"

    a.settings_schema = {"lol": {"type": "string"}, "li": {"type": "number"}}
    with pytest.raises(Exception):
        a.settings_schema = {
            "lol": {"type": "string"},
            "li": {"type": "number"},
            "required": ["li"],
        }


@pytest.mark.asyncio
async def test_basics_async():
    a = App("testkey", session="async")

    await (await a.owner).update(name="Myname2")
    assert await (await a.owner).name == "Myname2"

    assert len(await a.objects()) == 0

    o = await a.objects.create("myobj2", {"schema": {"type": "number"}})
    assert o.name == "myobj2"
    assert o.type == "timeseries"
    assert len(await a.objects()) == 1

    assert o == await a.objects[o.id]

    await o.delete()

    # assert len(await (await a.owner).apps())==1



import pytest
from heedy import App

# These tests are of bugfixes to heedy, making sure that things are working


def test_appscope():
    # Makes sure that removing owner's access doesn't affect the App's access
    a = App("testkey")
    o = a.objects.create("myobj")
    assert o.access == "*"
    o.key = "lol"
    o.owner_scopes = "read"
    assert o.read()["key"] == "lol"
    o.key = "hiya"
    assert o.read()["key"] == "hiya"


def test_metamod():
    a = App("testkey")
    o = a.objects.create("myobj", otype="timeseries")
    o.meta = {"schema": {"type": "number"}}
    assert o.cached_data["meta"]["schema"]["type"] == "number"
    assert o.cached_data["meta"]["actor"] == False

    assert o.meta["schema"]["type"] == "number"
    assert o.meta["actor"] == False

    with pytest.raises(Exception):
        o.meta = {"foo": "bar"}
